from configobj import ConfigObj
from functools import reduce
import io
import operator
import os
import pkg_resources
import validate


def flatten_validation_errors(errors):
    """Yields (path, error) pairs."""
    for section, value in errors.items():
        if value is True:
            continue
        elif value is False:
            yield section, 'value is missing'
        elif isinstance(value, dict):
            for path, errs in flatten_validation_errors(value):
                yield section + '.' + path, errs
        else:
            yield section, str(value)


class ValidationError(Exception):
    pass


class Config(object):
    """pip2nix configuration.

    This object handles merging and validation of CLI and .ini options."""

    def __init__(self):
        confspec = pkg_resources.resource_string(__name__, 'confspec.ini')
        self.config = ConfigObj(
            {}, configspec=io.StringIO(confspec.decode('utf-8')),
        )

    def __getitem__(self, key):
        return self.config[key]

    def get(self, key, default=None):
        try:
            self[key]
        except KeyError:
            return default

    def validate(self):
        """Check if configuration is OK, and raise a ValidationError if not."""
        validator = self._build_validator()
        errs = self.config.validate(validator, preserve_errors=True)
        if errs is not True:
            flat_errors = flatten_validation_errors(errs)
            err_msg = '\n'.join(
                path + ': ' + path_error
                for path, path_error in flat_errors)
            raise ValidationError(err_msg)

    def _build_validator(self):
        """Create a Validator with our custom rules included"""
        validator = validate.Validator()
        validator.functions['requirements_list'] = requirements_list_validator
        return validator

    def find_and_load(self):
        """Find a configuration file and load options from it."""
        base_path = os.getcwd()
        # Going up from CWD, find the first configuration file with [pip2nix*]
        while base_path != '/':
            path = os.path.join(base_path, 'pip2nix.ini')
            if os.path.exists(path):
                # Check if pip2nix sections exist in the file
                if self.load(path):
                    return
            base_path = os.path.dirname(base_path)

    def load(self, path):
        """Load configuration from path."""
        config = ConfigObj(path)
        if any(k == 'pip2nix' or k.startswith('pip2nix:') for k in config):
            # Only merge configuration files that actually support pip2nix
            self.merge_options(config.dict())
            return True
        return False

    def merge_options(self, options):
        # Expand sections with :
        for name, value in options.items():
            if ':' in name:
                opts = {}
                subopts = opts
                for elem in name.split(':'):
                    subopts[elem] = {}
                    last_subopts = subopts
                    subopts = subopts[elem]
                last_subopts[elem] = value
                self.merge_options(opts)
        self.config.merge(options)

    def merge_cli_options(self, cli_options):
        """Prepare the options before merging."""
        options = {}
        requirements = list(cli_options.get('specifiers', ()))
        requirements.extend(
            '-e ' + req for req in cli_options.get('editables', ()))
        requirements.extend(
            '-r ' + req for req in cli_options.get('requirements', ()))
        if requirements:
            options['requirements'] = requirements

        constraints = cli_options.get('constraints', ())
        if constraints:
            options['constraints'] = constraints

        for key in ('index_url', 'extra_index_url', 'no_index', 'output',
                    'licenses', 'only_direct'):
            try:
                value = cli_options[key]
                if value is not None:
                    options[key] = value
            except KeyError:
                pass

        self.merge_options({'pip2nix': options})

    def get_constraints(self):
        return self['pip2nix']['constraints']

    def get_requirements(self):
        """Yields pairs of (type, requirement) for all requirements.

        type is one of None, '-e', '-r'.
        """
        for req in self['pip2nix']['requirements']:
            if req.startswith(('-e', '-r')):
                yield req[:2], req[2:].strip()
            else:
                yield None, req.strip()

    def get_indexes(self):
        c = self['pip2nix']
        if c['no_index']:
            return []
        return list(filter(
            None, [c['index_url']] + c['extra_index_url']
        ))

    def get_package_config(self, package):
        """Get configuration for given package pair."""
        return self.get_config('pip2nix', 'package', package)

    def get_config(self, *path):
        try:
            return reduce(operator.getitem, path, self)
        except (KeyError, IndexError):
            return None


def requirements_list_validator(value, **kwargs):
    value = validate.force_list(value, **kwargs)
    validate.is_string_list(value)
    # TODO: check the formatting here?
    return value



import click
import pkg_resources
from .config import Config


@click.group()
def cli():
    pass


@cli.command()
@click.option('--build', '-b', type=click.Path(), metavar='<dir>',
              help="Directory to unpack packages and build in.")
@click.option('--download', '-d', type=click.Path(), metavar='<dir>',
              help="Directory to download packages to.")
# TODO:
#@click.option('--pre/--no-pre',
#              help="Also look for pre-release and unstable versions.")
@click.option('--output', metavar='<path>',
              help="Write the generated nix to <path>.")
@click.option('--only-direct', is_flag=True, default=None,
              help="Only render direct dependencies.")
@click.option('--index-url', '-i', metavar='<url>',
              help="Base URL of Python Package Index.")
@click.option('--extra-index-url', metavar='<url>',
              help="Extra index URLs to use.")
@click.option('--no-index/--index',
              help="Ignore indexes.")
#@click.option('--find-links', '-f', metavar='<url>',
#              help="Path or url to a package listing/directory.")

#TODO:
# --allow-external <package>  Allow the installation of a package even if it is externally hosted
# --allow-all-external        Allow the installation of all packages that are externally hosted
# --allow-unverified <package>
# Allow the installation of a package even if it is hosted in an insecure and unverifiable way
# --process-dependency-links  Enable the processing of dependency links.

@click.option('--configuration', metavar='<path>',
              help="Read pip2nix configuration from <path>.")

@click.option('--editable', '-e', multiple=True, type=click.Path(),
              metavar='<spec>',
              help="Add a requirement specifier (for pip install compatibility).")
@click.option('--requirement', '-r', multiple=True, type=click.Path(),
              metavar='<file>',
              help="Load specifiers from a requirements file.")
@click.option('--constraint', '-c', multiple=True, type=click.Path(),
              metavar='<file>',
              help="Constrain versions based on the constraints file.")
@click.option('--licenses/--no-licenses', default=False,
              help="Extract license information as well, off by default.")
@click.argument('specifiers', nargs=-1)
def generate(specifiers, **kwargs):
    """Generate a .nix file with specified packages."""
    kwargs['specifiers'] = specifiers + kwargs.pop('editable', [])
    kwargs['requirements'] = kwargs.pop('requirement', None)
    kwargs['constraints'] = kwargs.pop('constraint', None)
    kwargs['build_dir'] = kwargs.pop('build')
    kwargs['download_dir'] = kwargs.pop('download')

    config = Config()
    if kwargs['configuration']:
        config.load(kwargs['configuration'])
    else:
        config.find_and_load()
    config.merge_cli_options(kwargs)
    config.validate()

    from pip2nix.main import main
    from pip2nix.generate import generate
    import sys
    generate(config)


@cli.command()
@click.option('--configuration', metavar='<path>',
              help="Read pip2nix configuration from <path>.")
@click.option('--output', metavar='<path>', default='default.nix',
              help="Write the generated file to <path>.")
@click.option('--overrides-output', metavar='<path>',
              default='python-packages-overrides.nix',
              help="Write the generated overrides file to <path>.")
@click.option('--package', metavar='<package>',
              required=True,
              help="Name of the package the scaffold is for.")
def scaffold(output, overrides_output, **kwargs):
    import pip2nix

    config = Config()
    if kwargs['configuration']:
        config.load(kwargs['configuration'])
    else:
        config.find_and_load()
    config.merge_cli_options(kwargs)
    # TODO: Config enforces requirements to be specified, find a nicer
    # way to let Config know that we don't need requirements here.
    config.merge_options({'pip2nix': {'requirements': []}})
    config.validate()

    import jinja2
    raw_template = pkg_resources.resource_string(__name__, 'default.nix.j2')
    t = jinja2.Template(raw_template.decode('utf-8'))
    with open(output, 'w') as f:
        f.write(t.render(
            package_name=kwargs['package'],
            pip2nix_version=pip2nix.__version__))

    overrides_template = pkg_resources.resource_string(
        __name__, 'python-packages-overrides.nix.j2')
    t = jinja2.Template(overrides_template.decode('utf-8'))
    with open(overrides_output, 'w') as f:
        f.write(t.render(
            package_name=kwargs['package'],
            pip2nix_version=pip2nix.__version__))



from pip.req import RequirementSet


class RequirementSetLayer(RequirementSet):
    def __init__(self, *args, **kwargs):
        self.base_requirement_set = kwargs.pop('base')
        kwargs.setdefault('build_dir', self.base_requirement_set.build_dir)
        kwargs.setdefault('src_dir', self.base_requirement_set.src_dir)
        kwargs.setdefault('download_dir', self.base_requirement_set.download_dir)
        kwargs.setdefault('upgrade', self.base_requirement_set.upgrade)
        kwargs.setdefault('as_egg', self.base_requirement_set.as_egg)
        kwargs.setdefault('ignore_installed', self.base_requirement_set.ignore_installed)
        kwargs.setdefault('ignore_dependencies', self.base_requirement_set.ignore_dependencies)
        kwargs.setdefault('force_reinstall', self.base_requirement_set.force_reinstall)
        kwargs.setdefault('use_user_site', self.base_requirement_set.use_user_site)
        kwargs.setdefault('target_dir', self.base_requirement_set.target_dir)
        kwargs.setdefault('session', self.base_requirement_set.session)
        kwargs.setdefault('pycompile', self.base_requirement_set.pycompile)
        kwargs.setdefault('isolated', self.base_requirement_set.isolated)
        kwargs.setdefault('wheel_cache', self.base_requirement_set._wheel_cache)
        super(RequirementSetLayer, self).__init__(*args, **kwargs)

    def _prepare_file(
            self, finder, req_to_install, require_hashes=False,
            ignore_dependencies=False):
        if self.base_requirement_set.has_requirement(req_to_install.name):
            base_req = self.base_requirement_set.requirements[req_to_install.name]
            base_pkg_info = base_req.pkg_info()
            if not req_to_install.specifier.contains(base_pkg_info['Version']):
                # TODO: exceptions
                raise AssertionError(
                    ('There is already {req_name}=={base_pkg_version} downloaded, '
                     'but {comes_from} requires {req_name}{req_spec}')
                    .format(
                        req_name=req_to_install.name,
                        req_spec=req_to_install.specifier,
                        base_pkg_version=base_pkg_info['Version'],
                        comes_from=req_to_install.comes_from.name))
            return []
        else:
            extras = super(RequirementSetLayer, self) \
                ._prepare_file(
                    finder, req_to_install, require_hashes=require_hashes,
                    ignore_dependencies=ignore_dependencies)
            return extras



from __future__ import unicode_literals

from collections import defaultdict
from contexter import Contexter, contextmanager
from tempfile import mkdtemp
from itertools import chain
from operator import attrgetter
import os
import shutil

import pip
from pip import cmdoptions
from pip.utils.build import BuildDirectory
from pip.req import InstallRequirement
from pip.wheel import WheelCache
from pip.req import RequirementSet

import pip2nix
from .models.package import PythonPackage, indent
from .models.requirement_set import RequirementSetLayer


flatten = chain.from_iterable


@contextmanager
def temp_dir(name):
    path = mkdtemp('pip2nix')
    yield path
    shutil.rmtree(path)


class NixFreezeCommand(pip.commands.InstallCommand):

    name = 'pip2nix'
    usage = pip.commands.InstallCommand.usage.replace('%prog', name)
    summary = "Generate Nix expressions from requirements."

    PASSED_THROUGH_OPTIONS = (
        '--editable',
        '--requirement',
        '--build',
        '--download',
        '--download-cache',
        '--src',
        '--pre',
    )

    def __init__(self, pip2nix_config, *args, **kwargs):
        super(NixFreezeCommand, self).__init__(*args, **kwargs)
        self.config = pip2nix_config
        cmd_opts = self.cmd_opts
        for opt in cmd_opts.option_list:
            if opt.get_opt_string() not in self.PASSED_THROUGH_OPTIONS:
                cmd_opts.remove_option(opt.get_opt_string())

        cmd_opts.add_option('--configuration', metavar='CONFIG',
                            help="Read pip2nix configuration from CONFIG")
        cmd_opts.add_option('--output', metavar='OUTPUT',
                            help="Write the generated Nix to OUTPUT")

    def process_requirements(self, options, requirement_set, finder):
        top_levels = [r for r in requirement_set.requirements.values()
                      if r.comes_from is None]
        if self.config.get_config('pip2nix', 'only_direct'):
            packages_base = [
                r for r in requirement_set.requirements.values()
                if r.is_direct and not r.comes_from.startswith('-c ')]
        else:
            packages_base = requirement_set.requirements.values()
        packages = {
            req.name: PythonPackage.from_requirements(
                req, requirement_set._dependencies[req])
            for req in packages_base
            if not req.constraint
        }

        devDeps = defaultdict(list)
        with BuildDirectory(options.build_dir, delete=True):
            for req in top_levels:
                raw_tests_require = req.egg_info_data('tests_require.txt')
                if not raw_tests_require:
                    continue
                packages[req.name].check = True
                test_req_lines = filter(None, raw_tests_require.splitlines())
                for test_req_line in test_req_lines:
                    test_req = InstallRequirement.from_line(test_req_line)
                    devDeps[req.name].append(test_req)

            # TODO: this should be per-package
            # see https://github.com/ktosiek/pip2nix/issues/1#issuecomment-113716703
            test_req_set = self.get_tests_requirements_set(
                requirement_set, finder, flatten(devDeps.values()))

            for k, reqs in devDeps.items():
                test_deps = []
                for d in reqs:
                    if requirement_set.has_requirement(d.name):
                        test_deps.append((d.name, None))
                    else:
                        d.get_dist()
                        test_deps.append((d.name, d.pkg_info()['Version']))
                packages[k].test_dependencies = test_deps

            test_packages = {
                req.name: PythonPackage.from_requirements(
                    req, test_req_set._dependencies[req])
                for req in test_req_set.requirements.values()
                if not requirement_set.has_requirement(req.name)
            }

            for package in chain(packages.values(), test_packages.values()):
                pkg_config = self.config.get_package_config(package.name)
                if pkg_config:
                    package.override(pkg_config)

            include_lic = self.config['pip2nix']['licenses']

            with open(self.config['pip2nix']['output'], 'w') as f:
                self._write_about_comment(f)
                f.write('{ pkgs, fetchurl, fetchgit, fetchhg }:\n\n')
                f.write('self: super: {\n')
                f.write('  ' + indent(2, '\n'.join(
                    '"{}" = {}'.format(pkg.name,
                                     pkg.to_nix(include_lic=include_lic))
                    for pkg in sorted(packages.values(),
                                      key=attrgetter('name'))
                )))

                f.write('\n\n### Test requirements\n\n')
                f.write('  ' + indent(2, '\n'.join(
                    '{} = {}'.format(pkg.name,
                                     pkg.to_nix(include_lic=include_lic))
                    for pkg in sorted(test_packages.values(),
                                      key=attrgetter('name'))
                )))

                f.write('\n}\n')

    def _write_about_comment(self, target):
        target.write('# Generated by pip2nix {}\n'.format(pip2nix.__version__))
        target.write('# See https://github.com/johbo/pip2nix\n\n')

    def get_tests_requirements_set(self, base_set, finder, test_dependencies):
        test_req_set = RequirementSetLayer(base=base_set)
        for dep in test_dependencies:
            test_req_set.add_requirement(dep)
        test_req_set.prepare_files(finder)
        return test_req_set

    def super_run(self, options, args):
        """Copy of relevant parts from InstallCommand's run()"""
        temp_target_dir = (self.cleanup << temp_dir('pip2nix-temp-target'))

        with self._build_session(options) as session:
            finder = self._build_package_finder(options, session)
            wheel_cache = WheelCache(options.cache_dir, options.format_control)
            with BuildDirectory(options.build_dir, delete=True) as build_dir:
                requirement_set = RequirementSet(
                    build_dir=build_dir,
                    src_dir=options.src_dir,
                    download_dir=options.download_dir,
                    upgrade=options.upgrade,
                    as_egg=options.as_egg,
                    ignore_installed=options.ignore_installed,
                    ignore_dependencies=options.ignore_dependencies,
                    force_reinstall=options.force_reinstall,
                    use_user_site=options.use_user_site,
                    target_dir=temp_target_dir,
                    session=session,
                    pycompile=options.compile,
                    isolated=options.isolated_mode,
                    wheel_cache=wheel_cache,
                )

                self.populate_requirement_set(
                    requirement_set, args, options, finder, session, self.name,
                    wheel_cache
                )

                requirement_set.prepare_files(finder)

                self.process_requirements(options, requirement_set, finder)

                requirement_set.cleanup_files()

        return requirement_set

    def run(self, options, _args):
        with Contexter() as ctx:
            self.cleanup = ctx
            options, args = self.prepare_options(options)
            requirement_set = self.super_run(options, args)
            return requirement_set

    def prepare_options(self, options):
        """Load configuration from self.config into pip options.

        Returns a (options, args) tuple."""
        for opt_name, path in [
                ('index_url', ('index_url', )),
                ('output', ('output', )),
                ('build', ('build', )),
                ('download', ('download',)),
                ('src', ('src',))]:
            value = self.config.get_config('pip2nix', *path)
            if value is not None:
                setattr(options, opt_name, value)

        requirements = defaultdict(list)
        for req_type, req in self.config.get_requirements():
            requirements[req_type].append(req)

        args = requirements[None]
        options.requirements = requirements['-r']

        options.no_install = True  # Download only
        options.upgrade = True  # Download all packages
        options.use_wheel = False  # We'll build the wheels ourself
        options.no_clean = True  # This is needed to access pkg_info later
        options.download_dir = self.config.get_config('pip2nix', 'download') \
            or (self.cleanup << temp_dir('pip2nix'))
        options.constraints = self.config.get_constraints()

        # TODO: What are those about/for?
        cmdoptions.resolve_wheel_no_use_binary(options)
        cmdoptions.check_install_build_global(options)

        options.ignore_installed = True
        src_dir = self.config.get_config('pip2nix', 'src')
        options.src_dir = os.path.abspath(src_dir) if src_dir else None

        return options, args


def generate(config):
    cmd = NixFreezeCommand(config)

    return cmd.main([])



import os


def write_arg(cmd, basename, filename):
    argname = os.path.splitext(basename)[0]
    value = getattr(cmd.distribution, argname, None)
    if value is not None:
        value = '\n'.join(value)+'\n'
    cmd.write_or_delete_file(argname, filename, value)



from __future__ import unicode_literals

from contextlib import contextmanager
from collections import defaultdict
from copy import deepcopy
from subprocess import check_output
import sys

from .generate import NixFreezeCommand


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    cmd = NixFreezeCommand()

    return cmd.main(args)


if __name__ == '__main__':
    sys.exit(main())



from os import path
from setuptools import setup
from setuptools.command.test import test as TestCommand
import sys


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst'), 'rb') as f:
    long_description = f.read().decode('utf-8')

VERSION = '0.7.0'

setup(
    name="pip2nix",
    version=VERSION,
    description='Generate Nix expressions for Python packages.',
    long_description=long_description,
    url="https://github.com/johbo/pip2nix",
    author="Tomasz Kontusz",
    author_email="tomasz.kontusz@gmail.com",
    license='GPLv3+',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='nix pip',

    install_requires=[
        'pip>=8,<10',
        'configobj>=5',
        'click',
        'contexter',
        'jinja2',
    ],
    tests_require=['pytest'],
    packages=['pip2nix', 'pip2nix.models'],
    package_data={'pip2nix': ['*.ini', '*.j2']},
    cmdclass={'test': PyTest},
    entry_points={
        "console_scripts": [
            "pip2nix=pip2nix.cli:cli",
            "pip2nix%s=pip2nix.cli:cli" % sys.version[:1],
            "pip2nix%s=pip2nix.cli:cli" % sys.version[:3],
        ],
        "egg_info.writers": [
            "tests_require.txt=pip2nix.egg_writer:write_arg"
        ]
    }
)



import json
import os
from subprocess import check_output, STDOUT


_nix_licenses = None


def get_nix_licenses():
    """
    Generate a map of known licenses based on `nixpkgs`.
    """
    global _nix_licenses

    if _nix_licenses is None:
        nix_licenses_json = check_output([
            'nix-instantiate', '--eval', '--expr',
            'with import <nixpkgs> { }; builtins.toJSON lib.licenses'])
        nix_licenses_json = nix_licenses_json.decode('utf-8')

        # Dictionary which contains the contents of nixpkgs.lib.licenses.
        _nix_licenses = json.loads(json.loads(nix_licenses_json))

        # Convert all values to lowercase.
        for entry in _nix_licenses.values():
            for key, value in entry.items():
                try:
                    entry[key] = value.lower()
                except AttributeError:
                    # Skip values which don't have a lower() function.
                    pass

    return _nix_licenses


# Mapping from license name in setup.py to attribute in nixpkgs.lib.licenses.
# TODO: Think about providing this from outside, maybe from a file.
case_sensitive_license_nix_map = {
    'Apache 2.0': 'asl20',
    'Apache License, Version 2.0': 'asl20',
    'Apache Software License': 'asl20',
    'BSD license': 'bsdOriginal',
    'BSD': 'bsdOriginal',
    'GNU GPL': 'gpl1',
    'GNU GPLv2 or any later version': 'gpl2Plus',
    'GNU General Public License (GPL)': 'gpl1',
    'GNU General Public License v2 or later (GPLv2+)': 'gpl2Plus',
    'GPL': 'gpl1',
    'GPLv2 or later': 'gpl2Plus',
    'GPLv2': 'gpl2',
    'GPLv3': 'gpl3',
    'LGPLv2.1 or later': 'lgpl21Plus',
    'PSF License': 'psfl',
    'PSF': 'psfl',
    'Python Software Foundation License': 'psfl',
    'Python style': 'psfl',
    'Two-clause BSD license': 'bsd2',
    'ZPL 2.1': 'zpl21',
    'ZPL': 'zpl21',
    'Zope Public License': 'zpl21',
}
license_nix_map = {name.lower(): nix_attr
                   for name, nix_attr in
                   case_sensitive_license_nix_map.items()}


def indent(amount, string):
    lines = string.splitlines()
    if len(lines) == 0:
        return ''
    elif len(lines) == 1:
        return lines[0]
    else:
        return (
            lines[0] + '\n' +
            '\n'.join(' ' * amount + l for l in lines[1:]))


class PythonPackage(object):
    def __init__(self, name, version, dependencies, source, pip_req):
        """
        :param dependencies: list of (name, version) pairs.
        """
        self.name = name
        self.version = version
        self.dependencies = dependencies
        self.raw_args = {}
        self.source = source
        self.check = False
        self.test_dependencies = None
        self.pip_req = pip_req

    @classmethod
    def from_requirements(cls, req, deps):
        pkg_info = req.pkg_info()

        def name_version(dep):
            return (
                dep.name,
                dep.pkg_info()['Version'] if dep.source_dir else None,
            )

        return cls(
            name=req.name,
            version=pkg_info['Version'],
            dependencies=[name_version(d) for d in deps],
            source=req.link,
            pip_req=req,
        )

    def override(self, config):
        self.raw_args = config.get('args', {})

    def to_nix(self, include_lic):
        template = '\n'.join((
            'super.buildPythonPackage {{',
            '  {args}',
            '}};',
        ))
        meta_template = '\n'.join((
            'meta = {{',
            '  {meta_args}',
            '}};',
        ))

        args = dict(
            name='"{s.name}-{s.version}"'.format(s=self),
            doCheck='true' if self.check else 'false',
            src=link_to_nix(self.source),
        )

        if self.dependencies:
            args.update(dict(
                propagatedBuildInputs='[\n  ' + (
                    '\n  '.join('self."{}"'.format(name) for name, version
                                in self.dependencies)) + '\n]'
            ))

        if self.test_dependencies:
            args.update(dict(
                buildInputs='[\n  ' + (
                    '\n  '.join('self."{}"'.format(name) for name, version
                            in self.test_dependencies or ())) + '\n]'
            ))

        args.update(self.raw_args)

        # Prepare meta arguments.
        meta_args = dict()
        if include_lic:
            license_nix = self.get_license_nix()
            if license_nix:
                meta_args['license'] = license_nix

        # Render name first
        raw_args = 'name = {};'.format(args.pop('name'))
        for k, v in sorted(args.items()):
            raw_args += '\n{} = {};'.format(k, v)

        # Render meta arguments.
        if meta_args:
            raw_meta_args = ''
            for k, v in sorted(meta_args.items()):
                raw_meta_args += '{} = {};\n'.format(k, v)
            meta = meta_template.format(meta_args=indent(2, raw_meta_args))
            raw_args += '\n{}'.format(meta)

        return template.format(args=indent(2, raw_args))

    def get_license_nix(self):
        licenses = self.get_licenses_from_pkginfo()

        # Convert license strings to nix.
        nix_licenses = set()
        for lic in licenses:
            nix_licenses.add(license_to_nix(lic))

        template = '[ {licenses} ]'
        return template.format(licenses=' '.join(nix_licenses))

    def get_licenses_from_pkginfo(self):
        """
        Parses the license string from PKG-INFO file.
        """
        licenses = set()
        data = self.pip_req.egg_info_data('PKG-INFO')

        for line in data.split('\n'):

            # License string from setup() function.
            if line.startswith('License: '):
                lic = line.split('License: ')[-1]
                licenses.add(lic.strip())

            # License strings from classifiers.
            elif line.startswith('Classifier: License ::'):
                lic = line.split('::')[-1]
                licenses.add(lic.strip())

        return filter_licenses(licenses)


def filter_licenses(licenses):
    exclude = set(['UNKNOWN'])
    return licenses - exclude


def license_to_nix(license_name, nixpkgs='pkgs'):
    template = '{nixpkgs}.lib.licenses.{attribute}'
    full_name_template = '{{ fullName = "{full_name}"; }}'

    # Convert to lowercase for searching.
    full_name = license_name
    license_name = license_name.lower()

    # First try to fetch nix attribute name from custom mapping.
    attr = license_nix_map.get(license_name)
    if attr:
        return template.format(nixpkgs=nixpkgs, attribute=attr)

    # Otherwise try to look it up in the nix licenses.
    for attr, nix_license_data in get_nix_licenses().items():
        if license_name in nix_license_data.values():
            return template.format(nixpkgs=nixpkgs, attribute=attr)

    # No luck converting the license name to an attribute in
    # nixpkgs.lib.licenses. In this case we can at least store a set with the
    # fullName attribute like sets in nixpkgs.lib.licenses.
    return full_name_template.format(full_name=full_name)


def link_to_nix(link):
    if link.scheme == 'file':
        return './' + os.path.relpath(link.path)
    elif link.scheme in ('http', 'https'):
        print('Prefetching {url}.'.format(url=link.url_without_fragment))
        hash = prefetch_url(link.url_without_fragment)
        return '\n'.join((
            'fetchurl {{',
            '  url = "{url}";',
            '  {hash_name} = "{hash}";',
            '}}'
        )).format(
            url=link.url.split('#', 1)[0],
            hash=hash,
            hash_name='sha256',
        )
    elif link.scheme.startswith('git+'):
        url = link.url[len('git+'):]
        url = url.split('#', 1)[0]
        url, branch = url.rsplit('@', 1)
        print('Prefetching {url} at revision {revision}.'.format(
            url=url,
            revision=branch))
        hash, revision = prefetch_git(url, branch)
        return '\n'.join((
            'fetchgit {{',
            '  url = "{url}";',
            '  rev = "{revision}";',
            '  sha256 = "{hash}";',
            '}}',
        )).format(
            url=url,
            revision=revision,
            hash=hash,
        )
    elif link.scheme.startswith('hg+'):
        url = link.url[len('hg+'):]
        url = url.split('#', 1)[0]
        try:
            url, branch = url.rsplit('@', 1)
        except ValueError:
            branch = 'default'
        print('Prefetching {url} at revision {revision}.'.format(
            url=url,
            revision=branch))
        hash, revision = prefetch_hg(url, branch)
        return '\n'.join((
            'fetchhg {{',
            '  url = "{url}";',
            '  rev = "{revision}";',
            '  sha256 = "{hash}";',
            '}}',
        )).format(
            url=url,
            revision=revision,
            hash=hash,
        )
    else:
        raise NotImplementedError(
            'Unknown link scheme "{}"'.format(link.scheme))


def prefetch_git(url, rev):
    if len(rev) == 40 and rev.isdigit():
        rev_args = ['--rev', rev]
    else:
        rev_args = ['--branch-name', rev]
    out = check_output(['nix-prefetch-git'] + rev_args + [url])
    data = json.loads(out.decode('utf-8'))
    return data['sha256'], data['rev']


def prefetch_hg(url, rev):
    out = check_output(['nix-prefetch-hg', url, rev], stderr=STDOUT)
    data = {}
    for line in out.decode('utf-8').splitlines():
        if line.startswith('hash is '):
            data['sha256'] = line[len('hash is '):]
        if line.startswith('hg revision is '):
            data['rev'] = line[len('hg revision is '):]
    return data['sha256'], data['rev']


def prefetch_url(url):
    out = check_output(['nix-prefetch-url', url])
    data = out.decode('utf-8').strip()
    return data
